%
%
%  Connect Tracks Function for Imaris 7.3.0
%
%  Copyright Bitplane AG 2011
%
%
%  Installation:
%
%  - Copy this file into the XTensions folder in the Imaris installation directory.
%  - You will find this function in the Image Processing menu
%
%    <CustomTools>
%      <Menu>
%       <Submenu name="Tracks Functions">
%        <Item name="Connect Tracks" icon="Matlab" tooltip="Build bifurcated tracks.">
%          <Command>MatlabXT::XTTrackConnect(%i)</Command>
%        </Item>
%       </Submenu>
%      </Menu>
%      <SurpassTab>
%        <SurpassComponent name="bpSpots">
%          <Item name="Connect Tracks">
%            <Command>MatlabXT::XTTrackConnect(%i)</Command>
%          </Item>
%        </SurpassComponent>
%        <SurpassComponent name="bpSurfaces">
%          <Item name="Connect Tracks">
%            <Command>MatlabXT::XTTrackConnect(%i)</Command>
%          </Item>
%        </SurpassComponent>
%      </SurpassTab>
%    </CustomTools>
% 
%
%  Description:
%   
%   Build bifurcated tracks, connecting tracks start points to other
%     track points. 
%   Assumes linear motion of the objects to find best matches. 
%   
%

function XTTrackConnect(aImarisApplicationID)

% connect to Imaris interface
if ~isa(aImarisApplicationID, 'Imaris.IApplicationPrxHelper')
  javaaddpath ImarisLib.jar
  vImarisLib = ImarisLib;
  if ischar(aImarisApplicationID)
    aImarisApplicationID = round(str2double(aImarisApplicationID));
  end
  vImarisApplication = vImarisLib.GetApplication(aImarisApplicationID);
else
  vImarisApplication = aImarisApplicationID;
end

% the user has to create a scene with some surpass components
vSurpassScene = vImarisApplication.GetSurpassScene;
if isequal(vSurpassScene, [])
  msgbox('Please create some tracks in the surpass scene!');
  return
end

% get the selected object (spots or surfaces)
% read coordinates, time points and number of objects
vFactory = vImarisApplication.GetFactory;
vObjects = vImarisApplication.GetSurpassSelection;
if vFactory.IsSpots(vObjects)
  vObjects = vFactory.ToSpots(vObjects);
  vCoords = vObjects.GetPositionsXYZ;
  vTimes = vObjects.GetIndicesT + 1;
  vNumberOfObjects = numel(vTimes);
elseif vFactory.IsSurfaces(vObjects)
  vObjects = vFactory.ToSurfaces(vObjects);
  vNumberOfObjects = vObjects.GetNumberOfSurfaces;
  vCoords = zeros(vNumberOfObjects, 3);
  vTimes = ones(vNumberOfObjects, 1);
  for vIndex = 1:vNumberOfObjects
    vCoords(vIndex, :) = vObjects.GetCenterOfMass(vIndex - 1);
    vTimes(vIndex) = vObjects.GetTimeIndex(vIndex - 1) + 1;
  end
else
  msgbox('Please select some spots or surfaces!')
  return
end

% get the edges
vEdges = vObjects.GetTrackEdges + 1; % indices start from 1 here (matlab)
if isempty(vEdges)  
  msgbox('Please select some tracks!')
  return
end
vNumberOfEdges = size(vEdges, 1);

vAnswer = inputdlg('Please enter the maximal distance:', '', 1, {'5'});
if isempty(vAnswer), return, end
vMaxDistanceSquared = str2double(char(vAnswer))^2;

% expected position of the branching point in case of linear motion
vExpectedCoords = vCoords;

% get the beginning tracks
vIsBeginning = false(vNumberOfObjects, 1);
vAlreadyFound = false(vNumberOfObjects, 1);
for vIndex = 1:vNumberOfEdges
  
  % test left object
  vObject = vEdges(vIndex, 1);
  vOther = vEdges(vIndex, 2);
  % must be in the list only one time
  % its neighbor must have greater time point
  % must not be at first time point
  if vAlreadyFound(vObject)
    vIsBeginning(vObject) = false;
  elseif vTimes(vOther) > vTimes(vObject) && vTimes(vObject) ~= 1
    vIsBeginning(vObject) = true;
    % linear motion: calculate expected position of the previous point, 
    % i.e. coords(object) - (coords(next) - coords(object))
    vExpectedCoords(vObject, :) = 2*vCoords(vObject, :) - vCoords(vOther, :);
  end
  vAlreadyFound(vObject) = true;
  
  % test right object (just invert vObject and vOther)
  vObject = vEdges(vIndex, 2);
  vOther = vEdges(vIndex, 1);

  if vAlreadyFound(vObject)
    vIsBeginning(vObject) = false;
  elseif vTimes(vOther) > vTimes(vObject) && vTimes(vObject) ~= 1
    vIsBeginning(vObject) = true;
    vExpectedCoords(vObject, :) = 2*vCoords(vObject, :) - vCoords(vOther, :);
  end
  vAlreadyFound(vObject) = true;
  
end
vBeginningPoints = find(vIsBeginning);

% find best candidate as previous of beginning points
vNumberOfBeginningPoints = numel(vBeginningPoints);
vAdditionalEdges = zeros(vNumberOfBeginningPoints, 2);
vBeginningPointHasMatch = false(vNumberOfBeginningPoints, 1);
for vIndex = 1:vNumberOfBeginningPoints
  vObject = vBeginningPoints(vIndex);
  vObjectCoords = vExpectedCoords(vObject, :);
  vObjectTime = vTimes(vObject);
  
  % possible matches are all the points at vObjectTime - 1
  vOthers = find(vTimes == vObjectTime - 1);
  if ~isempty(vOthers)
    vBestIndex = vOthers(1);
    vBestDistanceSquared = sum((vObjectCoords - vCoords(vBestIndex, :)).^2);
    for vOtherIndex = 2:numel(vOthers)
      vOther = vOthers(vOtherIndex);
      vDistanceSquared = sum((vObjectCoords - vCoords(vOther, :)).^2);
      if vDistanceSquared < vBestDistanceSquared
        vBestIndex = vOther;
        vBestDistanceSquared = vDistanceSquared;
      end
    end
    if (vBestDistanceSquared < vMaxDistanceSquared) 
      vAdditionalEdges(vIndex, :) = [vObject, vBestIndex];
      vBeginningPointHasMatch(vIndex) = true;
    end
  end
end

% discard invalid pairs
vAdditionalEdges = vAdditionalEdges(vBeginningPointHasMatch, :);

% finally modify imaris component
vObjects.SetTrackEdges([vEdges; vAdditionalEdges] - 1);


